from typing import Any, Dict, Union

from aiohttp.web_urldispatcher import DynamicResource
from jj.matchers import AllMatcher as JJAllMatcher
from jj.matchers import AnyMatcher as JJAnyMatcher
from jj.matchers import EqualMatcher as JJEqualMatcher
from jj.matchers import MethodMatcher as JJMethodMatcher
from jj.matchers import PathMatcher as JJPathMatcher
from jj.matchers import ResolvableMatcher
from jj.matchers.attribute_matchers import RouteMatcher as JJRouteMatcher

from ._common import destroy_prefix, normalize_path

__all__ = ('create_openapi_matcher', )


class BaseMatcher:
    def match(self, spec_unit: tuple[str, str]) -> bool:
        raise NotImplementedError()


class MethodMatcher(BaseMatcher):
    def __init__(self, mocked_method: Any) -> None:
        self._mocked_method = mocked_method

    def match(self, spec_unit: tuple[str, str]) -> bool:
        return bool(self._mocked_method == spec_unit[0])

    def __repr__(self) -> str:
        """
        Return a string representation of the MethodMatcher instance.

        :return: A string describing the class and matcher.
        """
        return f"{self.__class__.__qualname__}({self._mocked_method!r})"


class _Resource(DynamicResource):
    def match(self, path: str) -> Union[Dict[str, str], None]:
        return self._match(path)


class RouteMatcher(BaseMatcher):
    def __init__(self, mocked_path: str) -> None:
        self._mocked_path = mocked_path

    def match(self, spec_unit: tuple[str, str]) -> bool:
        mock = self._mocked_path
        spec = spec_unit[1]

        if "{" in mock:
            return normalize_path(mock) == normalize_path(spec)
        else:
            self._resource_spec = _Resource(spec)
            return self._resource_spec.match(mock) is not None

    def __repr__(self) -> str:
        """
        Return a string representation of the RouteMatcher instance.

        :return: A string describing the class and matcher.
        """
        return f"{self.__class__.__qualname__}({self._mocked_path!r})"


class AnyMatcher(BaseMatcher):
    def __init__(self, matchers: list[BaseMatcher]) -> None:
        assert len(matchers) > 0
        self._matchers = matchers

    def match(self, spec_unit: tuple[str, str]) -> bool:
        for matcher in self._matchers:
            if matcher.match(spec_unit):
                return True
        return False

    def __repr__(self) -> str:
        """
        Return a string representation of the AnyMatcher instance.

        :return: A string describing the class and matchers.
        """
        return (f"{self.__class__.__qualname__}"
                f"({self._matchers!r}")


class AllMatcher(BaseMatcher):
    def __init__(self, matchers: list[BaseMatcher]) -> None:
        assert len(matchers) > 0
        self._matchers = matchers

    def match(self, spec_unit: tuple[str, str]) -> bool:
        for matcher in self._matchers:
            if not matcher.match(spec_unit):
                return False
        return True

    def __repr__(self) -> str:
        """
        Return a string representation of the AllMatcher instance.

        :return: A string describing the class and matchers.
        """
        return (f"{self.__class__.__qualname__}"
                f"({self._matchers!r})")


def create_openapi_matcher(matcher: ResolvableMatcher, prefix: str | None = None) -> BaseMatcher | None:
    spec_matcher: BaseMatcher

    if isinstance(matcher, JJMethodMatcher):
        submatcher = matcher.sub_matcher
        if isinstance(submatcher, JJEqualMatcher):
            spec_matcher = MethodMatcher(mocked_method=submatcher.expected)
            return spec_matcher
        else:
            return None

    elif isinstance(matcher, JJPathMatcher):
        submatcher = matcher.sub_matcher
        if isinstance(submatcher, JJRouteMatcher):
            mocked_path = submatcher.path
            if prefix:
                mocked_path = destroy_prefix(submatcher.path, prefix)
            spec_matcher = RouteMatcher(mocked_path=mocked_path)
            return spec_matcher
        else:
            return None

    elif isinstance(matcher, JJAllMatcher):
        submatchers = matcher.sub_matchers
        matchers = [create_openapi_matcher(submatcher, prefix=prefix) for submatcher in submatchers if create_openapi_matcher(submatcher, prefix=prefix) is not None]
        spec_matcher = AllMatcher(matchers=matchers)
        return spec_matcher

    elif isinstance(matcher, JJAnyMatcher):
        submatchers = matcher.sub_matchers
        matchers = [create_openapi_matcher(submatcher, prefix=prefix) for submatcher in submatchers if create_openapi_matcher(submatcher, prefix=prefix) is not None]
        spec_matcher = AnyMatcher(matchers=matchers)
        return spec_matcher
    else:
        return None



import asyncio
from functools import wraps
from typing import Callable, TypeVar

from jj import RelayResponse

from ._config import Config
from .validator import Validator

_T = TypeVar('_T')


def validate_spec(*,
                  spec_link: str | None,
                  skip_if_failed_to_get_spec: bool | None = None,
                  is_raise_error: bool | None = None,
                  is_strict: bool | None = None,
                  prefix: str | None = None,
                  force_strict: bool = False,
                  ) -> Callable[[Callable[..., _T]], Callable[..., _T]]:
    """
    Validates the jj mock function with given specification lint.

    Args:
       skip_if_failed_to_get_spec: If True - skip validation if failed to get spec. False is default.
       spec_link: The link to the specification. `None` for disable validation.
       is_raise_error: If True - raises error when validation is failes. False is default.
       is_strict: If True - validate exact structure in given mocked.
       prefix: Prefix is used to cut paths prefix in mock function.
       force_strict: If True - forced remove all Ellipsis from the spec.
    """
    def decorator(func: Callable[..., _T]) -> Callable[..., _T]:
        func_name = func.__name__

        validator = Validator(
            spec_link=spec_link,
            prefix=prefix,
            func_name=func_name,
            force_strict=force_strict,
            skip_if_failed_to_get_spec=skip_if_failed_to_get_spec if skip_if_failed_to_get_spec is not None else Config.SKIP_IF_FAILED_TO_GET_SPEC,
            is_raise_error=is_raise_error if is_raise_error is not None else Config.IS_RAISES,
            is_strict=is_strict if is_strict is not None else Config.IS_STRICT
            )

        @wraps(func)
        async def async_wrapper(*args: object, **kwargs: object) -> _T:
            mocked = await func(*args, **kwargs)
            if validator.spec_link:
                if isinstance(mocked.handler.response, RelayResponse):
                    print("RelayResponse type is not supported")
                    return mocked
                validator.validate(mocked)
            else:...
            return mocked

        @wraps(func)
        def sync_wrapper(*args: object, **kwargs: object) -> _T:
            mocked = func(*args, **kwargs)
            if validator.spec_link:
                if isinstance(mocked.handler.response, RelayResponse):
                    print("RelayResponse type is not supported")
                    return mocked
                validator.validate(mocked)
            else:...
            return mocked

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper
    return decorator



class Config:
    # service
    MAIN_DIRECTORY = "spec_validator"
    GET_SPEC_TIMEOUT = 30.0

    # interface
    OUTPUT_FUNCTION = None  # can be used for custom output func

    # params
    IS_RAISES = False
    IS_STRICT = False
    SKIP_IF_FAILED_TO_GET_SPEC = False



import json
from hashlib import md5
from os import makedirs, path, remove
from pickle import dump
from pickle import load as pickle_load
from time import time
from typing import Any, Dict, List, Tuple

import httpx
from schemax_openapi import SchemaData, collect_schema_data
from yaml import CLoader, load

from .._config import Config
from ..validator_base import BaseValidator

__all__ = ('load_cache', )

CACHE_DIR = Config.MAIN_DIRECTORY + '/_cache_parsed_specs'
CACHE_TTL = 3600  # in second


def _build_entity_dict(entities: List[SchemaData]) -> Dict[Tuple[str, str], SchemaData]:
    entity_dict = {}
    for entity in entities:
        entity_key = (entity.http_method.upper(), entity.path)
        entity_dict[entity_key] = entity
    return entity_dict


def _validate_cache_file(filename: str) -> bool:
    if not path.isfile(filename):
        return False

    file_age = time() - path.getmtime(filename)

    if file_age > CACHE_TTL:
        remove(filename)
        return False

    return True


def _get_cache_filename(url: str) -> str:
    hash_obj = md5(url.encode())
    return path.join(CACHE_DIR, hash_obj.hexdigest() + '.cache' + '.yml')


def _download_spec(validator: BaseValidator) -> httpx.Response | None:
    if validator.skip_if_failed_to_get_spec:
        try:
            response = httpx.get(validator.spec_link, timeout=Config.GET_SPEC_TIMEOUT)
        except httpx.ConnectTimeout as e:
            validator.output(e, f"Timeout occurred while trying to connect to the {validator.spec_link}.")
            return None
        except httpx.ReadTimeout as e:
            validator.output(e, f"Timeout occurred while trying to read the spec from the {validator.spec_link}.")
            return None
        except httpx.HTTPError as e:
            validator.output(e, f"An error occurred while trying to download the spec: {e}")
            return None
        except Exception as e:
            validator.output(e, f"An error occurred while trying to download the spec: {e}")
            return None
        response.raise_for_status()
        return response
    else:
        try:
            response = httpx.get(validator.spec_link, timeout=Config.GET_SPEC_TIMEOUT)
        except httpx.ConnectTimeout:
            raise httpx.ConnectTimeout(f"Timeout occurred while trying to connect to the {validator.spec_link}.")
        except httpx.ReadTimeout:
            raise httpx.ReadTimeout(f"Timeout occurred while trying to read the spec from the {validator.spec_link}.")
        except httpx.HTTPError as e:
            raise httpx.HTTPError(f"An error occurred while trying to download the spec: {e}")
        except Exception as e:
            raise ValueError(f"An error occurred while trying to download the spec: {e}")
        response.raise_for_status()
        return response


def _save_cache(spec_link: str, raw_schema: dict[str, Any]) -> None:
    filename = _get_cache_filename(spec_link)
    makedirs(CACHE_DIR, exist_ok=True)
    with open(filename, 'wb') as f:
        dump(raw_schema, f)


def load_cache(validator: BaseValidator) -> Dict[Tuple[str, str], SchemaData] | None:
    filename = _get_cache_filename(validator.spec_link)

    if _validate_cache_file(filename):
        with open(filename, 'rb') as f:
            raw_schema = pickle_load(f)
    else:
        raw_spec = _download_spec(validator)
        if raw_spec is None:
            return None

        content_type = raw_spec.headers.get('Content-Type', '')

        if 'application/json' in content_type:
            raw_schema = json.loads(raw_spec.text)
        elif 'text/yaml' in content_type or 'application/x-yaml' in content_type:
            raw_schema = load(raw_spec.text, Loader=CLoader)
        else:
            # trying to match via file extension
            if validator.spec_link.endswith('.json'):
                raw_schema = json.loads(raw_spec.text)
            elif validator.spec_link.endswith('.yaml') or validator.spec_link.endswith('.yml'):
                raw_schema = load(raw_spec.text, Loader=CLoader)
            else:
                raise ValueError(f"Unsupported content type: {content_type}")

        _save_cache(validator.spec_link, raw_schema)

    parsed_data = collect_schema_data(raw_schema)
    prepared_dict = _build_entity_dict(parsed_data)

    return prepared_dict



from ._cacheir import load_cache
from ._common import destroy_prefix, normalize_path, validate_non_strict
from ._refiner import get_forced_strict_spec
from ._spec_matcher import create_openapi_matcher

__all__ = ('load_cache', 'destroy_prefix', 'normalize_path', 'validate_non_strict', 'get_forced_strict_spec', 'create_openapi_matcher')


import re
from typing import Any, Match

from district42 import GenericSchema
from revolt import SubstitutorValidator
from valera import ValidationException, format_result

__all__ = ('normalize_path', 'destroy_prefix', 'validate_non_strict', )


def normalize_path(path: str) -> str:
    var_counter = 1

    def replace_var(match: Match[str]) -> str:
        nonlocal var_counter
        replacement = f'{{var{var_counter}}}'
        var_counter += 1
        return replacement

    normalized_path = re.sub(r'{[a-zA-Z0-9_]+}', replace_var, path)

    return normalized_path

def destroy_prefix(path: str, prefix: str) -> str:
    return re.sub(prefix, '', path)

def validate_non_strict(schema: GenericSchema, value: dict[str, Any]) -> bool:
    result = schema.__accept__(SubstitutorValidator(), value=value)
    if errors := format_result(result):
        raise ValidationException("\n".join(errors))
    return True



from abc import ABC, abstractmethod
from typing import Any


class BaseValidator(ABC):
    @abstractmethod
    def validate(self, mocked: Any) -> None:
        pass

    @abstractmethod
    def prepare_data(self) -> dict:
        pass

    @abstractmethod
    def output(self, e: Exception, text: str = None) -> None:
        pass

    @property
    @abstractmethod
    def func_name(self) -> str:
        pass

    @property
    @abstractmethod
    def spec_link(self) -> str | None:
        pass

    @property
    @abstractmethod
    def skip_if_failed_to_get_spec(self) -> bool:
        pass


from district42 import GenericSchema
from district42.types import AnySchema, DictSchema, ListSchema
from district42.utils import is_ellipsis
from niltype import Nil

__all__ = ('get_forced_strict_spec', )


def get_forced_strict_spec(schema: GenericSchema) -> GenericSchema:
    if isinstance(schema, DictSchema):
        if schema.props.keys is not Nil:
            new_keys = {}
            for k, (v, is_optional) in schema.props.keys.items():
                if not is_ellipsis(k):
                    new_keys[k] = (get_forced_strict_spec(v), is_optional)
            return schema.__class__(schema.props.update(keys=new_keys))
        return schema
    elif isinstance(schema, ListSchema):
        if schema.props.elements is not Nil:
            new_elements = [get_forced_strict_spec(element) for element in schema.props.elements]
            return schema.__class__(schema.props.update(elements=new_elements))
        elif schema.props.type is not Nil:
            new_type = get_forced_strict_spec(schema.props.type)
            return schema.__class__(schema.props.update(type=new_type))
        return schema
    elif isinstance(schema, AnySchema):
        if schema.props.types is not Nil:
            new_types = tuple(get_forced_strict_spec(t) for t in schema.props.types)
            return schema.__class__(schema.props.update(types=new_types))
        return schema
    else:
        return schema
